///////////////////////////////////////////////////////
// BLE Slave Sensor Readout Example                  //
// User Application                                  //
// This example is the simplest RF communication     //
// Example. By setting #define ROLE, you can compile //
// it for sender or receiver                         //
///////////////////////////////////////////////////////

#include <Wire.h>
#include <InertialSensor.h>
#include <BRIX2.h>
#include <BGLib.h>

// ================================================================
// BLE STATE TRACKING (UNIVERSAL TO JUST ABOUT ANY BLE PROJECT)
// ================================================================

// BLE state machine definitions
#define BLE_STATE_STANDBY           0
#define BLE_STATE_SCANNING          1
#define BLE_STATE_ADVERTISING       2
#define BLE_STATE_CONNECTING        3
#define BLE_STATE_CONNECTED_MASTER  4
#define BLE_STATE_CONNECTED_SLAVE   5

// BLE state/link status tracker
uint8_t ble_state = BLE_STATE_STANDBY;


#define BLE_WAKEUP_PIN MISO // Assuming Analog Input 1 (PF6) is connected to BLE wake-up pin (P1_6)
#define BLE_RESET_PIN   9   // BLE reset pin (active-low)

#define GATT_HANDLE_BATTERY 11
#define GATT_HANDLE_ACC 11
#define GATT_HANDLE_GYRO 15
#define GATT_HANDLE_LED 19


uint8 data_acc[] = { 	0x00, 0x00, // X
						0x00, 0x00, // Y
						0x00, 0x00, // Z
};
uint8 data_gyr[] = { 	0x00, 0x00, // X
						0x00, 0x00, // Y
						0x00, 0x00, // Z
};

uint8 data_battery[] = { 0x00};


// create BGLib object:
//  - use SoftwareSerial por for module comms
//  - use nothing for passthrough comms (0 = null pointer)
//  - enable packet mode on API protocol since flow control is unavailable
BGLib ble113((HardwareSerial *) &Serial1, 0, 1);
#define BGAPI_GET_RESPONSE(v, dType) dType *v = (dType *)ble113.getLastRXPayload()


BRIX2 myBrick;
InertialSensor motionSensor;

bool led_user_controlled = false;

bool notification_battery = false;
bool notification_acc = false;
bool notification_gyro = false;

bool loopdata = false;

bool first_start = true;

// ================================================================
// ARDUINO APPLICATION SETUP AND LOOP FUNCTIONS
// ================================================================

#define SERIAL_STATUS
void setup() {



	// Start Serial1 with 38400 bps (to SysController), turn RXTX off, stop Serial1 and start Serial1 to BLE113
	Serial1.begin(38400); //this is absolutely required in order to use the system functions!
	myBrick.RXTXOff();
	Serial1.end();

	myBrick.initialize();
	motionSensor.initialize();

	// BRIX --> PC
	Serial.begin(115200);

	setupBGLib();

	// BRIX --> BLE113
	// Attention: You need to adapt the baudrate to the ble113 firmware baudrate  in hardware.xml
	Serial1.begin(230400);

}

// The loop function is called in an endless loop
void loop() {
	ble113.ble_cmd_gap_end_procedure();
	delay(1000);
	startAdvertising();

	while (1) {

		// keep polling for new data from BLE
		ble113.checkActivity();

		readAccData();
		readGyroData();

		if (ble_state == BLE_STATE_STANDBY && !led_user_controlled) {
			myBrick.setRGB(255, 0, 0);
		} else if (ble_state == BLE_STATE_ADVERTISING && !led_user_controlled) {
			myBrick.setRGB(0, 0, 255);
		} else if (ble_state == BLE_STATE_CONNECTED_SLAVE
				&& !led_user_controlled) {
			myBrick.setRGB(0, 255, 0);
		}
	}
}


void readAccData() {
	if (notification_acc) {
		int ax = motionSensor.getAccelerationX();
		data_acc[0] = ax;
		data_acc[1] = ax >> 8;
		int ay = motionSensor.getAccelerationY();
		data_acc[2] = ay;
		data_acc[3] = ay >> 8;
		int az = motionSensor.getAccelerationZ();
		data_acc[4] = az;
		data_acc[5] = az >> 8;
		ble113.ble_cmd_attributes_write(GATT_HANDLE_ACC, 0, 6, data_acc);
		delay(20);
	}
}
void readGyroData() {
	if (notification_gyro) {
		int gx = motionSensor.getRotationX();
		data_gyr[0] = gx;
		data_gyr[1] = gx >> 8;
		int gy = motionSensor.getRotationY();
		data_gyr[2] = gy;
		data_gyr[3] = gy >> 8;
		int gz = motionSensor.getRotationZ();
		data_gyr[4] = gz;
		data_gyr[5] = gz >> 8;
		ble113.ble_cmd_attributes_write(GATT_HANDLE_GYRO, 0, 6, data_gyr);
		delay(20);
	}
}

void startAdvertising() {
    // set advertisement interval to 200-300ms, use all advertisement channels
    // (note min/max parameters are in units of 625 uSec)
    ble113.ble_cmd_gap_set_adv_parameters(320, 480, 7);
    while (ble113.checkActivity(1000));

    // start advertising general discoverable / undirected connectable
    ble113.ble_cmd_gap_set_mode(BGLIB_GAP_GENERAL_DISCOVERABLE, BGLIB_GAP_UNDIRECTED_CONNECTABLE);
    while (ble113.checkActivity(1000));

    // set state to ADVERTISING
    ble_state = BLE_STATE_ADVERTISING;

}

void setupBGLib() {
	// set up internal status handlers (these are technically optional)
	ble113.onBusy = onBusy;
	ble113.onIdle = onIdle;
	ble113.onTimeout = onTimeout;

	// ONLY enable these if you are using the <wakeup_pin> parameter in your firmware's hardware.xml file
	// BLE module must be woken up before sending any UART data
	//ble113.onBeforeTXCommand = onBeforeTXCommand;
	//ble113.onTXCommandComplete = onTXCommandComplete;

    // set up BGLib event handlers
    ble113.ble_evt_system_boot = my_ble_evt_system_boot;
    ble113.ble_evt_connection_status = my_ble_evt_connection_status;
    ble113.ble_evt_connection_disconnected = my_ble_evt_connection_disconnect;
    ble113.ble_evt_attributes_value = my_ble_evt_attributes_value;
    ble113.ble_evt_attributes_status = my_ble_evt_attributes_status;

    ble113.ble_rsp_gap_set_scan_parameters = my_rsp_gap_set_scan_parameters;
    ble113.ble_rsp_gap_set_mode = my_rsp_gap_set_mode;
    ble113.ble_rsp_attributes_write = my_rsp_attributes_write;
}

// ================================================================
// INTERNAL BGLIB CLASS CALLBACK FUNCTIONS
// ================================================================

// called when the module begins sending a command
void onBusy() {
    // turn LED on when we're busy
    //digitalWrite(LED_PIN, HIGH);
}

// called when the module receives a complete response or "system_boot" event
void onIdle() {
    // turn LED off when we're no longer busy
    //digitalWrite(LED_PIN, LOW);
}

// called when the parser does not read the expected response in the specified time limit
void onTimeout() {
    // reset module (might be a bit drastic for a timeout condition though)
    digitalWrite(BLE_RESET_PIN, LOW);
    delay(5); // wait 5ms
    digitalWrite(BLE_RESET_PIN, HIGH);
}

// called immediately before beginning UART TX of a command
void onBeforeTXCommand() {
    // wake module up (assuming here that digital pin 5 is connected to the BLE wake-up pin)
    //digitalWrite(BLE_WAKEUP_PIN, HIGH);

    // wait for "hardware_io_port_status" event to come through, and parse it (and otherwise ignore it)
    uint8_t *last;
    while (1) {
        ble113.checkActivity();
        last = ble113.getLastEvent();
        if (last[0] == 0x07 && last[1] == 0x00) break;
    }

    // give a bit of a gap between parsing the wake-up event and allowing the command to go out
    delayMicroseconds(1000);
}

// called immediately after finishing UART TX
void onTXCommandComplete() {
    // allow module to return to sleep (assuming here that digital pin 5 is connected to the BLE wake-up pin)
    //digitalWrite(BLE_WAKEUP_PIN, LOW);
}


// ================================================================
// APPLICATION EVENT HANDLER FUNCTIONS
// ================================================================
void my_rsp_gap_set_mode(const ble_msg_gap_set_mode_rsp_t *msg) {
	if (msg->result != 0 ) {
		led_user_controlled = true;
		myBrick.setRGB(255,255,255);
	}
}
void my_rsp_gap_set_scan_parameters(const ble_msg_gap_set_scan_parameters_rsp_t *msg) {
#ifdef SERIAL_STATUS
	Serial.print("<--\tgap_set_scan_parameters: { ");
	Serial.print("result: "); Serial.print((uint16_t)msg -> result, HEX);
	Serial.println(" }");
#endif
}

void my_rsp_attributes_write(const struct ble_msg_attributes_write_rsp_t * msg) {
#ifdef SERIAL_STATUS
	Serial.print("- attributes_write: { ");
	Serial.print("result: "); Serial.print((uint16_t)msg -> result, HEX);
	Serial.println(" }");
#endif

}

void my_ble_evt_system_boot(const ble_msg_system_boot_evt_t *msg) {
    #ifdef SERIAL_STATUS
        Serial.print("###\tsystem_boot: { ");
        Serial.print("major: "); Serial.print(msg -> major, HEX);
        Serial.print(", minor: "); Serial.print(msg -> minor, HEX);
        Serial.print(", patch: "); Serial.print(msg -> patch, HEX);
        Serial.print(", build: "); Serial.print(msg -> build, HEX);
        Serial.print(", ll_version: "); Serial.print(msg -> ll_version, HEX);
        Serial.print(", protocol_version: "); Serial.print(msg -> protocol_version, HEX);
        Serial.print(", hw: "); Serial.print(msg -> hw, HEX);
        Serial.println(" }");
    #endif

        startAdvertising();


}

void my_ble_evt_connection_status(const ble_msg_connection_status_evt_t *msg) {
    #ifdef SERIAL_STATUS
        Serial.print("###\tconnection_status: { ");
        Serial.print("connection: "); Serial.print(msg -> connection, HEX);
        Serial.print(", flags: "); Serial.print(msg -> flags, HEX);
        Serial.print(", address: ");
        // this is a "bd_addr" data type, which is a 6-byte uint8_t array
        for (uint8_t i = 0; i < 6; i++) {
            if (msg -> address.addr[i] < 16) Serial.write('0');
            Serial.print(msg -> address.addr[i], HEX);
        }
        Serial.print(", address_type: "); Serial.print(msg -> address_type, HEX);
        Serial.print(", conn_interval: "); Serial.print(msg -> conn_interval, HEX);
        Serial.print(", timeout: "); Serial.print(msg -> timeout, HEX);
        Serial.print(", latency: "); Serial.print(msg -> latency, HEX);
        Serial.print(", bonding: "); Serial.print(msg -> bonding, HEX);
        Serial.println(" }");
    #endif

    // check for new connection established
	if ((msg->flags & 0x05) == 0x05) {
		// track state change based on last known state, since we can connect two ways
		if (ble_state == BLE_STATE_ADVERTISING) {
			ble_state = BLE_STATE_CONNECTED_SLAVE;
		}
	}
}

void my_ble_evt_connection_disconnect(const struct ble_msg_connection_disconnected_evt_t *msg) {
#ifdef SERIAL_STATUS
	Serial.print("###\tconnection_disconnect: { ");
	Serial.print("connection: "); Serial.print(msg -> connection, HEX);
	Serial.print(", reason: "); Serial.print(msg -> reason, HEX);
	Serial.println(" }");
#endif

	// After a disconnect, start advertising again
	startAdvertising();
}

void my_ble_evt_attributes_status(const struct ble_msg_attributes_status_evt_t * msg) {
#ifdef SERIAL_STATUS
	Serial.print("### Attribute Status Event ");
	Serial.print("handle: "); Serial.print(msg -> handle, HEX);
	Serial.print(", flags: "); Serial.print(msg -> flags, HEX);
	Serial.println(" }");
#endif


	switch (msg->handle) {
	case GATT_HANDLE_ACC:
		if (msg->flags == 1) {
			notification_acc = true;
		} else {
			notification_acc = false;
		}
		break;
	case GATT_HANDLE_GYRO:
		if (msg->flags == 1) {
			notification_gyro = true;
		} else {
			notification_gyro = false;
		}
		break;
	}

	if (msg->flags != 0) {  // Notifications activated
		loopdata = true;
	} else {
		loopdata = false;
	}
}

void my_ble_evt_attributes_value(const struct ble_msg_attributes_value_evt_t *msg) {
    #ifdef SERIAL_STATUS
        Serial.print("###\tattributes_value: { ");
        Serial.print("connection: "); Serial.print(msg -> connection, HEX);
        Serial.print(", reason: "); Serial.print(msg -> reason, HEX);
        Serial.print(", handle: "); Serial.print(msg -> handle, HEX);
        Serial.print(", offset: "); Serial.print(msg -> offset, HEX);
        Serial.print(", value_len: "); Serial.print(msg -> value.len, HEX);
        Serial.print(", value_data: ");
        // this is a "uint8array" data type, which is a length byte and a uint8_t* pointer
        for (uint8_t i = 0; i < msg -> value.len; i++) {
            if (msg -> value.data[i] < 16) Serial.write('0');
            Serial.print(msg -> value.data[i], HEX);
        }
        Serial.println(" }");
    #endif


    	if (msg->handle == GATT_HANDLE_LED) {
    		uint8 r = msg->value.data[0];
    		uint8 g = msg->value.data[1];
    		uint8 b = msg->value.data[2];
    		led_user_controlled = true;
    		myBrick.setRGB(r,g,b);
    	}
}
